*
*  Tiny-Forth runtime executive for the 68hc16
*
*  This version must be used with tiny4th v3.0 or greater.  It will
*  not work with earlier versions of tiny4th.
*
*  Restrictions imposed on the 68hc16 by tiny4th:
*
*  1.  All tiny4th code, including the runtime executive, must reside
*  in bank 0.  All tiny4th variables and timers must reside in bank 0.
*
*  2.  The return stack will always be placed in the SRAM array.  This
*  array will be moved to the address assigned to stkbeg by the user at
*  compile-time.
*


pitvec	equ	$00070			vector for PIT interrupts

start
	ldab	#$0f			initialize EK to bank F
	tbek				extended addressing reaches I/O regs

	ldaa	#$7f			initialize system clock
	staa	syncr			set to 16.78 MHz
	clr	sypcr			turn off COP (on after reset)

*
*  Locate the SRAM at the requested address.  The hardware stack
*  (S) and the data stack (Y) will always reside in SRAM.  SK and
*  YK are both set to reach into the bank holding SRAM.
*

        ldd     #(stkbeg/$10000)&$0f	calc bank to hold SRAM
        std     rambah			store high ram array bank number
	tbsk				set SK to the proper bank
	ldd	#(stkbeg&$fc00)		calc low addr of 1K SRAM
        std     rambal			store low ram array
        clr     rammcr			enable ram
        lds     #(stkbeg&$fffe)		put SP at top of 1k internal SRAM
	ldd	#(datastk/$10000)&$0f	calc page for data stack
	tbyk				set YK to the proper bank
	ldy	#(datastk&$fffe)	set top of data stack

*
*  Set up the Periodic Interrupt Timer (PIT) to generate an interrupt
*  every 4.148 msecs.  This is backward compatible with the 68hc11
*  version of tiny4th.  The user's code may subsequently alter this
*  timing value if necessary.
*

	ldd	#$0438			req level 4, vector #$38
	std	picr			set up periodic interrupt CR
	ldd	#$0022			PIT period = 4.148 msec
	std	pitr			set timer register

	clrb				get a 0 for EK
	tbek				extended addressing reaches bank 0
	ldd	#pitisr			get address of PIT ISR
	std	pitvec			write to vector addr
	andp	#$ff7f			mask interrupts below level 4

*
*  Start up the tiny4th run-time executive.
*

	ldx	#firstop		get addr of first opcode
	stx	ip			save as instruction ptr
	jmp	next			and start it up


*
*  star		multiply nos by tos, leave 16-bit product
*

star
	jsr	pulld			get tos to d
	tde				move to e
	ldd	2,y			get nos to d
	emuls				do signed multiply
	std	2,y			save low 16 bits on stack
	bra	next			and outta here


*
*  /mod		divide nos by tos, return rem (nos) and quotient (tos)
*

slmod
	bsr	dodiv			divide nos by tos
	std	4,y			rem to nos
	stx	2,y			quot to tos
	bra	next			and leave


*
*  /		divide nos by tos, return quotient
*

slash
	bsr	dodiv			divide nos by tos
	aiy	#2			drop top item
	stx	2,y			save quotient
	bra	next			and leave

*
*  mod		divide nos by tos, return modulo
*

mod
	bsr	dodiv			divide nos by tos
	aiy	#2			drop top item
	std	2,y			save remainder
	bra	next			and leave


dodiv
	ldd	4,y			get nos to d
	ldx	2,y			get tos to x
	idiv				calc d/x
	rts

*
*  dup		duplicate top item on stack
*

dup
	ldd	2,y			get tos
	bra	pushdn			and push a copy onto stack


*
*  ?dup		duplicate top item on stack if it is not 0
*

qdup
	ldd	2,y			get tos
	beq	next			if 0, all done
	bra	pushdn			push a copy onto stack



*
*  swap		exchange tos with nos
*

swap
	ldd	2,y			get tos
	ldx	4,y			get nos
	stx	2,y			save to tos
	std	4,y			save to nos
	bra	next			and leave



*
*  rot		rearrange stack;  n3 n2 n1 -- n2 n1 n3
*

rot
	ldd	6,y			save n3
	ldx	4,y			get n2
	stx	6,y			n2 --> n3
	ldx	2,y			get n1
	stx	4,y			n1 --> n2
	std	2,y			n3 --> n1
	bra	next



*
*  over		dup nos to tos;  n2 n1 -- n2 n1 n2
*

over
	ldd	4,y			get nos
pushdn
	bsr	pushd			save to stack
*					and fall into next


next
	ldx	ip			get addr of next opcode
	ldab	0,x			get the code
	aslb				make it a word pointer
	inx				bump the ip
	stx	ip			and save it back
	ldx	#optable		point to op code table
	abx				calc index
	ldx	0,x			get the addr
	jmp	0,x			execute the word


pushx
	stx	0,y			put x onto stack
pushx1
	aiy	#-2			push stack pointer down
	rts

pushd
	std	0,y			put d onto stack
	bra	pushx1			and finish up





*
*  drop		pop top element from stack, don't save it
*

drop
	aiy	#2			move stack pointer up
	bra	next			and leave



pullx
	ldx	2,y			get tos into x
pullx1
	aiy	#2			move stack pointer up
	rts				and leave

pulld
	ldd	2,y			get tos into d
	bra	pullx1			and leave




*
*  !		store 16-bit value at addr
*

store
	bsr	pullx			pop addr into x
	bsr	pulld			pop data into d
	std	0,x			store it
	bra	next			and leave


*
*  c!		store 8-bit value at addr
*

cstore
	bsr	pullx			pop addr into x
	bsr	pulld			pop data into d
	stab	0,x			store it
	bra	next			and leave


*
*  @		get 16-bit value from addr
*

fetch
	ldx	2,y			get the addr
	ldd	0,x			get the contents
fetch1
	std	2,y			put on stack
	bra	next			and return


*
*  c@		get 8-bit value from addr
*

cfetch
	ldx	2,y			get the addr
	ldab	0,x			get the byte
	clra				leave top empty
	bra	fetch1			and leave




*
*  docon	fetch contents of ip and put on stack
*

docon
	bsr	ip2dw			fetch (ip)
	jmp	pushdn			push onto stack and leave


ip2dw
	ldx	ip			get addr of next code
	ldd	0,x			get the code
	inx				bump twice for word
ip2dw1
	inx				bump once for byte
	stx	ip			save new ip
	rts



*
*  doconb	fetch contents of ip (byte) and put on stack
*

doconb
	bsr	ip2db			get (ip) as byte
	jmp	pushdn			save on stack and leave



ip2db
	ldx	ip			get addr of next code
	ldab	0,x			get the code (byte)
	clra				zero top byte
	bra	ip2dw1			and leave



*
*  +		add 16-bit values on stack
*

plus
	bsr	pulld			get tos into d
plus1
	addd	2,y			add to tos
	bra	fetch1			and leave



*
*  negtos	Calculate -tos, return in d
*
*  This hidden routine copies tos into d, then negates it.  It leaves
*  the data stack intact.
*

negtos
	clra				get a big zero
	clrb
	subd	2,y			calc -tos
	rts					and leave



*
*  negate	Replace tos with -tos
*

negate
	bsr	negtos		calc -tos
	std	2,y			and save on stack
	jmp	next			and leave



*
*  -		subtract tos from nos
*

minus
	bsr	negtos			calc -tos
	aiy	#2			drop top item
	bra	plus1			and add -tos to nos



*
*  abs		Replace tos with abs(tos)
*

abs
	ldaa	2,y			test MSB of tos
	bmi	negate			if set, use negate
	jmp	next			not set, leave



*
*  do		initialize return stack for DO loops
*

do
	ldx	ip			get addr of next cell
	pshx				save on return stack
	ldd	4,y			get DO limit
	addd	#$8000			calc limit + $8000
	std	p1			temp save
	ldd	2,y			get DO index
	subd	p1			calc index - (limit + $8000)
	ldx	p1			get limit + $8000
	pshx				save adjusted limit
	pshb				save adjusted index
	psha	
	aiy	#4			pop two cells from stack
	bra	skip			and hop around next cell


*
*  +loop	increment and test for DO-+LOOP construct
*

ploop
	jsr	pulld			get increment value
	bra	loop1			and do it


*
*  loop		increment and test for DO-LOOP construct
*

loop
	ldd	#1			get a 1 for incrementing
loop1
	tsx				make x the return stack
	addd	0,x			add top item on return stack
	bvs	loops			branch if overflowed to $8000
	std	0,x			no, save back onto return stack
	bra	bran			need to loop back; use BRANCH
loops
	pulx				pull three words from return stack
	pulx
	pulx
	bra	skip			and skip next word


*
*  _i	 	return index of outer DO-LOOP
*

_i
	clrb				force b to 0
_i1
	tsx				use x as return stack pointer
	abx				add offset to proper loop
	ldd	0,x			get index
	addd	2,x			adjust for offset
	jmp	pushdn			save on data stack and leave



*
*  _j		return index of 2nd DO-LOOP
*

_j
	ldab	#6			get offset to 2nd loop
	bra	_i1			and reuse some code


*
*  _k		return index of 3rd DO-LOOP
*

_k
	ldab	#12			get offset to 3rd loop
	bra	_i1			and reuse some code


*
*  leave	exit current DO-LOOP
*

leave
	pulx				remove index
	pulx				remove limit
	pulx				get addr of ending offset
	stx	p1			save in working area
	ldd	0,x			now get offset
	addd	p1			calc addr of bottom of loop
	std	ip			save as new ip
	jmp	next			and leave loop



*
*  bran		branch (relative) to next thread
*

bran
	ldx	ip			get current ip
	ldd	0,x			get value at ip
	addd	ip			add to current ip
	std	ip			and make it new ip
	jmp	next			and leave


*
*  beq		branch (relative) to next thread if tos = 0
*

beq
	jsr	pulld			get tos
	cpd	#0			got a zero?
	beq	bran			if so, change ip to new value
skip
	jsr	ip2dw			no, eat next word in thread
	jmp	next			and leave	




*
*  docol	push current ip, use (ip) as new token thread
*

docol
	ldx	ip			get current ip
	ldd	0,x			get value at ip
	aix	#2			advance ip 2 bytes
	pshx				save on hardware stack
	std	ip			change to new thread
	jmp	next			and outta here


*
*  semi		pop ip from hardware stack, return to old thread
*

semi
	pulx				get old ip
	stx	ip			restore old ip
	jmp	next			and leave


*
*  execute	use word in tos as address of next thread
*

execute
	ldx	ip			get return addr
	pshx				save on stack
	jsr	pullx			get addr of next word
	stx	ip			save as next word
	jmp	next			and do the word

	


*
*  docode	jump to native code immediately after token
*
*  The native code must end with a JMP NEXT instruction.
*

docode
 	ldx	ip			get ip
	pshx				save for now
	aix	#2			ip + 2 = next token addr
	stx	ip			save for later
	pulx				get ip back
	ldx	0,x			now get address of native code
	jmp	0,x			and do native code




*
*  true		push true flag ($ffff) onto stack
*

truef
	ldd	#TRUE			get the flag
	jmp	pushdn			push and leave


*
*  false	push false flag (0) onto stack
*

falsef
	clra				get a big 0
	clrb
	jmp	pushdn			push and leave



*
*  and		logical and of tos and nos
*

and
	jsr	pulld			get tos into d
	anda	2,y			and msb
	andb	3,y			and lsb
and1
	std	2,y			put back on stack
	jmp	next			and leave


*
*  or		logical or of tos and nos
*

or
	jsr	pulld			get tos into d
	oraa	2,y			or msb
	orab	3,y			or lsb
	bra	and1			and finish up


*
*  xor		logical exclusive-or of tos and nos
*

xor
	jsr	pulld			get tos into d
	eora	2,y			xor msb
	eorb	3,y			xor lsb
	bra	and1			and finish up



*
*  not		negation of tos
*

not
	ldd	2,y			get tos into d
	coma				invert msb
	comb				invert lsb
	bra	and1			and finish up	


*
*  compset (hidden word); used by comparisons below
*

compset
	ldx	2,y			get tos to x
	iny				(local code for speed)
	iny
	ldd	2,y			get nos to d
	stx	2,y			save tos back for now
	cpd	2,y			compare nos to tos
	rts				and leave


*
*  =		true if nos = tos
*

equal
	bsr	compset			set up comparison
	beq	ttruef			branch if the same
ffalsef
	ldd	#FALSE			false, get the flag
	bra	and1			and leave
ttruef
	ldd	#TRUE			true, get the flag
	bra	and1			and leave


*
*  <		true if nos < tos
*

less
	bsr	compset			set up comparison
	blt	ttruef			leave true if nos < tos
	bra	ffalsef			else leave false


*
*  >		true if nos > tos
*

great
	bsr	compset			set up comparison
	bgt	ttruef			leave true if nos > tos
	bra	ffalsef			else leave false


*
*  U<		true if nos < tos (unsigned)
*

uless
	bsr	compset			set up comparison
	blo	ttruef			leave true if nos < tos (U)
	bra	ffalsef			else leave false


*
*  U>		true if nos > tos (unsigned)
*

ugreat
	bsr	compset			set up comparison
	bhi	ttruef			leave true if nos > tos (U)
	bra	ffalsef			else leave false


*
*  0=		invert logical flag in tos
*

zequal
	ldd	2,y			get tos into d
	beq	ttruef			if 0, make it true
	bra	ffalsef			else make it false



*
*  shiftr	shift nos right by count in tos
*

shiftr
	bsr	shiftset		set up for shifting
	cpx	#0			reached the end?
	beq	and1			if so, leave
shiftr1
	lsrd				no, move it
	dex				count this shift
	bne	shiftr1			loop until done
	bra	and1			and leave



*
*  shiftl	shift nos left by count in tos
*

shiftl
	bsr	shiftset		set up for shifting
	cpx	#0			reached the end?
	beq	and1			if so, leave
shiftl1
	lsld				no, move it
	dex				count this shift
	bne	shiftl1			loop until done
	lbra	and1			and leave



shiftset
	jsr	pulld			get count
	clra				force it to be <16
	andb	#$0f
	xgdx				move to x
	ldd	2,y			get value
	rts				and leave

*
*  +!		Add nos to (tos), save to (tos)
*

plusstr
	jsr	pullx			get addr
	jsr	pulld			get value
	addd	0,x			add value to current contents
	std	0,x			and save as new contents
	jmp	next			and outta here


*
*  1		Push a 1 onto stack
*

one
	ldd	#1			get a one
	jmp	pushdn			and outta here


*
*  2		Push a 2 onto stack
*

two
	ldd	#2			get a two
	jmp	pushdn			and leave
	




*
*  >r		move tos to return stack
*

tor
	jsr	pullx			get tos to x
	pshx				save on return stack
	jmp	next			and leave


*
*  r>		move return stack to tos
*

rfrom
	pulx				get top element on return stack
	xgdx				move to d
	jmp	pushdn			put on stack and leave


*
*  I/O access words.  The 68hc16 I/O registers are all stored in
*  bank F, which means they cannot be reached with normal @ or
*  C@ instructions.
*
*  The following low-level words use the F-bank equivalent of the
*  specified address to perform the @ or !.
*

*
*  _ioprep		Prepare XK:IX for I/O register transfers
*

_ioprep
	ldab	#$0f			get bank number for I/O regs
	tbxk				XK now points to bank F
	jmp	pullx			get the address and return
	

*
*  io@			Get a word from the I/O space
*

iofetch
	bsr	_ioprep			set up the X registers
	lde	0,x			get the word at TOS
	clrb				get bank number for base page
	tbxk				XK now points to bank 0
	ted				move data to D
	jmp	pushdn			and leave


*
*  io!			Store a word to the I/O space
*

iostore
	bsr	_ioprep			set up X
	jsr	pulld			get the data
	std	0,x			store in I/O area
iostore1
	clrb				get 0 for base bank
	tbxk				XK now point to bank 0
	jmp	next			and leave


*
*  ioc@			Fetch a byte from the I/O space
*

iocfetch
	bsr	_ioprep			set up X
	ldab	0,x			get byte from I/O area
	pshb				save in stack
	clrd				0s to A and B
	tbxk				XK now points at bank 0
	pulb				restore the byte
	jmp	pushdn			and leave


*
*  ioc!			Save a byte to the I/O space
*

iocstore
	bsr	_ioprep			set the X reg
	jsr	pulld			get the TOS
	stab	0,x			save the byte
	bra	iostore1		borrow some code



*
*  optable	list of addresses, based on token pointed to by IP.
*
*		routine			token and description
*		------			------------------------------
optable
	fdb	dup			00; duplicate tos
	fdb	swap			01; exchange tos with nos
	fdb	drop			02; pop tos and discard
	fdb	rot			03; n3 n2 n1 --> n1 n3 n2
	fdb	fetch			04; get word stored at addr in tos
	fdb	cfetch			05; get byte stored at addr in tos
	fdb	store			06; put (word) tos in addr at nos
	fdb	cstore			07; put (byte) tos in addr at nos
	fdb	bran			08; ip = ip + (ip)
	fdb	beq			09; bran if tos = 0
	fdb	docon			0a; put (ip) onto stack as word
	fdb	doconb			0b; put (ip) onto stack as byte
	fdb	docol			0c; jump to (ip), start new thread
	fdb	semi			0d; return to old thread
	fdb	truef			0e; push true flag onto stack
	fdb	falsef			0f; push false flag onto stack
	fdb	and			10; tos = tos & nos
	fdb	or			11; tos = tos | nos
	fdb	xor			12; tos = tos ^ nos
	fdb	not			13; tos = ~tos
	fdb	shiftr			14; shift nos right by count in tos
	fdb	shiftl			15; shift nos left by count in tos
	fdb	plus			16; tos = nos + tos
	fdb	minus			17; tos = nos - tos
	fdb	equal			18; nos = tos ?
	fdb	less			19; nos < tos ?
	fdb	great			1a; nos > tos ?
	fdb	uless			1b; nos < tos (U) ?
	fdb	ugreat			1c; nos > tos (U) ?
	fdb	zequal			1d; invert logical tos
	fdb	over			1e; dup nos to tos
	fdb	slmod			1f; nos/tos, return q and r
	fdb	mod			20; nos/tos, return r
	fdb	slash			21; nos/tos, return q
	fdb	star			22; nos * tos
	fdb	plusstr			23; nos + (tos) -> (tos)
	fdb	one			24; push 1 onto stack
	fdb	two			25; push 2 onto stack
	fdb	negate			26; tos = -tos
	fdb	abs			27; tos = abs(tos)
	fdb	qdup			28; dup tos if not 0
	fdb	do			29; DO run-time code
	fdb	loop			2a; LOOP run-time code
	fdb	ploop			2b; +LOOP run-time code
	fdb	_i			2c; I run-time code
	fdb	_j			2d; J run-time code
	fdb	_k			2e; K run-time code
	fdb	leave			2f; LEAVE current DO-LOOP
	fdb	tor			30; move tos to return stack
	fdb	rfrom			31; move return stack to tos
	fdb	docode			32; start executing CODE word
	fdb	execute			33; vectored execution
	fdb	iofetch			34; fetch from I/O space
	fdb	iostore			35; store to I/O space
	fdb	iocfetch		36; c-fetch from I/O space
	fdb	iocstore		37; c-store to I/O space



*
*  pitisr -- periodic interval timer interrupt service routine
*
*  This routine automatically checks all timers in the timers array.  If
*  any timer is not currently zero, it is decremented.  Using the power-up
*  default timing values for the 68hc16, this routine will execute once
*  every 4.148 msecs, assuming an 32 KHz crystal.
*

pitisr
	pshm	d,x,k			save some regs
	clrb				get a 0
	tbxk				force XK to point to bank 0
	ldx	#_time0			point x at timers
pit1
	cpx	#_time1			done all timers?
	beq	pitx			branch if so
	ldd	0,x			get next timer
	beq	pit2			branch if already 0
	subd	#1			drop this timer
	std	0,x			and save it back
pit2
	aix	#2			move to next word
	bra	pit1			do some more
pitx
	pulm	d,x,k			restore the regs
	rti				and outta here



